import ims
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import CenteredNorm
from sklearn.cross_decomposition import PLSRegression
from sklearn.metrics import mean_squared_error, r2_score
from ims.utils import vip_scores


class PLSR:
    """
    Applies a scikit-learn PLSRegression to GC-IMS data and provides prebuilt
    plots as well as a feature selection via variable importance in projection (VIP)
    scores.

    See the scikit-learn documentation for more details:
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSRegression.html

    Parameters
    ----------
    dataset : ims.Dataset
        Needed for the retention and drift time coordinates in the plots.

    n_components : int, optional
        Number of components to keep, by default 2.

    kwargs : optional
        Additional key word arguments are passed on to the scikit-learn PLSRegression.

    Attributes
    ----------
    x_scores : numpy.ndarray of shape (n_samples, n_components)
        X scores.

    y_scores : numpy.ndarray of shape (n_samples, n_components)
        y scores.

    x_weights : numpy.ndarray of shape (n_features, n_components)
        The left singular vectors of the cross-covariance matrices of each iteration.

    y_weights : numpy.ndarray of shape (n_targets, n_components)
        The right singular vectors of the cross-covariance matrices of each iteration.

    x_loadings : numpy.ndarray of shape (n_features, n_components)
        The loadings of X. When scaling was applied on the dataset,
        corrects the loadings using the weights.

    y_loadings : numpy.ndarray of shape (n_targes, n_components)
        The loadings of y.

    coefficients : numpy.ndarray of shape (n_features, n_targets)
        The coefficients of the linear model.

    y_pred_train : numpy.ndarray
        Stores the predicted values from the training data for the plot method.

    Example
    -------
    >>> import ims
    >>> import pandas as pd
    >>> ds = ims.Dataset.read_mea("IMS_data")
    >>> responses = pd.read_csv("responses.csv")
    >>> ds.labels = responses
    >>> X_train, X_test, y_train, y_test = ds.train_test_split()
    >>> model = ims.PLSR(ds, n_components=5)
    >>> model.fit(X_train, y_train)
    >>> model.predict(X_test, y_test)
    >>> model.plot()
    """

    def __init__(self, dataset, n_components=2, **kwargs):
        self.dataset = dataset
        self.n_components = n_components
        self.pls = PLSRegression(n_components=n_components, scale=False, **kwargs)
        self._fitted = False
        self._validated = False

    def fit(self, X_train, y_train):
        """
        Fits the model with training data.

        Parameters
        ----------
        X_train : numpy.ndarray of targets (n_samples, n_features)
            Training vectors with features.

        y_train : numpy.ndarray of shape (n_samples, n_targets)
            Target vectors with response variables.

        Returns
        -------
        self
            Fitted model.
        """
        self.pls.fit(X_train, y_train)
        self.x_scores, self.y_scores = self.pls.transform(X_train, y_train)
        self.x_weights = self.pls.x_weights_
        self.x_loadings = self.pls.x_loadings_
        self.y_weights = self.pls.y_weights_
        self.y_loadings = self.pls.y_loadings_
        self.coefficients = self.pls.coef_
        self.y_pred_train = self.pls.predict(X_train).flatten()
        self.y_train = y_train
        self.vip_scores = ims.utils.vip_scores(
            self.x_weights, self.x_scores, self.y_loadings
        )
        self.selectivity_ratio = ims.utils.selectivity_ratio(X_train, self.coefficients)
        self._fitted = True
        return self

    def predict(self, X_test, y_test=None):
        """
        Predicts responses for features of the test data.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Features of test data.

        y_train : numpy.ndarray of shape (n_samples, n_targets), optional
            True labels for test data.
            If set allows automatic plotting of validation data,
            by default None.

        Returns
        -------
        numpy.ndarray of shape (n_samples, n_targets)
            Predicted responses for test data.
        """
        y_pred = self.pls.predict(X_test)
        self.y_pred = y_pred
        self.y_test = y_test
        if y_test is not None:
            self.rmse = round(mean_squared_error(y_test, self.y_pred, squared=False), 2)
        self._validated = True
        return y_pred

    def score(self, X_test, y_test, sample_weight=None):
        """
        Calculates R^2 score score for predicted data.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Feature vectors of the test data.

        y_test : numpy.ndarray of shape (n_samples, n_targets)
            True regression responses.

        sample_weight : array-like of shape (n_samples,), optional
            Sample weights, by default None.

        Returns
        -------
        score : float
            R^2 score.
        """
        y_pred = self.predict(X_test)
        return r2_score(y_test, y_pred, sample_weight=sample_weight)

    def transform(self, X, y=None):
        """
        Apply the dimensionality reduction.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            Feature matrix.
        y : numpy.ndarray of shape (n_samples, n_targtets), optional
            Dependend variables, by default None

        Returns
        -------
        X_scores
        """
        return self.pls.transform(X, y)

    def plot(self, annotate=False):
        """
        Plots predicted vs actual values and shows regression line.
        Recommended to predict with test data first.

        annotate : bool, optional
            If True annotates plot with sample names,
            by default False.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        if self._validated and hasattr(self, "y_test"):
            ax = sns.scatterplot(x=self.y_train, y=self.y_pred_train, label="Training")
            sns.scatterplot(
                ax=ax,
                x=self.y_test,
                y=self.y_pred.flatten(),
                label=f"Validation: RMSE {self.rmse}",
            )
        else:
            ax = sns.scatterplot(x=self.y_train, y=self.y_pred_train, label="Training")

        z = np.polyfit(self.y_train, self.y_pred_train, 1)
        sns.lineplot(
            ax=ax,
            x=np.polyval(z, self.y_train),
            y=self.y_train,
            c="tab:blue",
            label="Trend line",
        )

        z = np.polyfit(self.y_train, self.y_train, 1)
        sns.lineplot(
            ax=ax,
            x=np.polyval(z, self.y_train),
            y=self.y_train,
            c="tab:red",
            linestyle=":",
            label="Perfect fit",
        )

        if annotate:
            if hasattr(self.dataset, "train_index"):
                sample_names = self.dataset[self.dataset.train_index].samples
                X = self.y_train
                Y = self.y_pred_train
            if self._validated and hasattr(self, "y_test"):
                sample_names = (
                    sample_names + self.dataset[self.dataset.test_index].samples
                )
                X = np.concatenate((X, self.y_test))
                Y = np.concatenate((Y, self.y_pred.flatten()))
            for x, y, sample in zip(X, Y, sample_names):
                plt.text(x, y, sample)

        plt.xlabel("Observed response")
        plt.ylabel("Predicted response")
        plt.legend()
        return ax

    def plot_loadings(self, component=1, color_range=0.01, width=6, height=6):
        """
        Plots PLS x loadings as image with retention and drift
        time coordinates.

        Parameters
        ----------
        component : int, optional
            Component to plot, by default 1.

        color_range : float, optional
            Minimum and maximum to adjust to different scaling methods,
            by default 0.02.

        width : int or float, optional
            Width of the plot in inches,
            by default 8.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        loadings = self.x_loadings[:, component - 1].reshape(self.dataset[0].shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            loadings,
            cmap="RdBu_r",
            vmin=(-color_range),
            vmax=color_range,
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time), min(ret_time), max(ret_time)),
        )

        plt.colorbar(label="PLS loadings")
        plt.title(f"PLS loadings of component {component}")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax

    def plot_coefficients(self, width=6, height=6):
        """
        Plots PLS coefficients as image with retention and drift time axis.

        Parameters
        ----------
        width : int or float, optional
            Width of the plot in inches,
            by default 8.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        coef = self.coefficients.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            coef,
            cmap="RdBu_r",
            norm=CenteredNorm(0),
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time), min(ret_time), max(ret_time)),
        )

        plt.colorbar(label="PLS coefficients")
        plt.title("PLS coefficients")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax

    def plot_vip_scores(self, threshold=None, width=6, height=6):
        """
        Plots VIP scores as image with retention and drift time axis.

        Parameters
        ----------
        threshold : int
            Only plots VIP scores above threshold if set.
            Values below are displayed as 0,
            by default None.

        width : int or float, optional
            Width of the plot in inches,
            by default 8.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.

        Returns
        -------
        matplotlib.pyplot.axes

        Raises
        ------
        ValueError
            If VIP scores have not been calculated prior.
        """
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        if threshold is None:
            vip_matrix = self.vip_scores.reshape(self.dataset[0].values.shape)
        else:
            vips = np.zeros_like(self.vip_scores)
            i = np.where(self.vip_scores > threshold)
            vips[i] = self.vip_scores[i]
            vip_matrix = vips.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            vip_matrix,
            cmap="RdBu_r",
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time), min(ret_time), max(ret_time)),
        )

        plt.colorbar(label="VIP scores")
        plt.title(f"PLS VIP scores")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax

    def plot_selectivity_ratio(self, threshold=None, width=6, height=6):
        """
        Plots VIP scores as image with retention and drift time axis.

        Parameters
        ----------
        threshold : int
            Only plots VIP scores above threshold if set.
            Values below are displayed as 0,
            by default None.

        width : int or float, optional
            Width of the plot in inches,
            by default 8.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        if threshold is None:
            sr_matrix = self.vip_scores.reshape(self.dataset[0].values.shape)
        else:
            sratio = np.zeros_like(self.vip_scores)
            i = np.where(self.vip_scores > threshold)
            sratio[i] = self.vip_scores[i]
            sr_matrix = sratio.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            sr_matrix,
            cmap="RdBu_r",
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time), min(ret_time), max(ret_time)),
        )

        plt.colorbar(label="Selectivity ratio")
        plt.title(f"PLS selectivity ratio")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax



#%%
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import seaborn as sns
from scipy.stats import f
from sklearn.decomposition import PCA
import scipy
import pandas as pd
from matplotlib.ticker import AutoMinorLocator

class OneClassSIMCA:
    """
    Implements soft independed modelling of class analogies (SIMCA) for one target class.

    Parameters
    ----------
    n_components : int, optional
        Number of PCA components, by default 2.

    q : float, optional
        q-value: false discovery rate, by default 0.95.

    Attributes
    ---------
    n_components : int
        Parameter set during initialization.

    q : float
        Parameter set during initialization.

    pca : sklearn.decomposition.PCA
        The underlying PCA model.

    target : str
        Name of the target class. Parameter set with 'fit' method.

    Q_target : numpy.ndarray of shape (n_samples,)
        Q residuals of target class. Calculated in 'fit' method.

    Q_conf : float
        Q confidence limit of target class. Calculated in 'fit' method.

    Q_test : numpy.ndarray of shape (n_samples,)
        Q residuals of test samples. Calculated in 'predict' method.

    Tsq_target : numpy.ndarray of shape (n_samples,)
        T square values of target class. Calculated in 'fit' method.

    Tsq_conf : float
        T square confidence limit of target class. Calculated in 'fit' method.

    Tsq_test : numpy.ndarray of shape (n_samples,)
        T square of test samples. Calculated in 'predict' method.

    Example
    -------
    >>> import numpy as np
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.utils import resample
    >>> from sklearn.metrics import accuracy_score
    >>>
    >>> # load example data and set target (class label)
    >>> X, y = load_iris(return_X_y=True)
    >>> target = 0
    >>>
    >>> # Find target data to train model and draw random test data
    >>> X_target = X[np.where(y == target)]
    >>> X_test, y_test = resample(X, y n_samples=50)
    >>>
    >>> # instantiate a model and fit target data
    >>> model = OneClassSIMCA(n_components=2, q=0.95)
    >>> model.fit(X_target, target)
    >>>
    >>> # make prediction on test data and calculate accuracy
    >>> y_pred = model.predict(X_test)
    >>> y_true = y_test == target
    >>> print(accuracy_score(y_true, y_pred))
    >>>
    >>> # visualize results
    >>> model.plot(hue=y_test)
    """

    def __init__(self, n_components=2, q=0.95):
        self.n_components = n_components
        self.q = q
        self.pca = PCA(n_components=n_components)
        self._fitted = False
        self._validated = False

    def fit(self, X_target, target):
        """
        Fit SIMCA model with data from target class.

        Parameters
        ----------
        X_target : numpy.ndarray of shape (n_samples, n_features)
            Training data for target class.

        target : str
            Name of the target class as identifier of the model and for plotting.
        """
        # set target as identifier and title in plot method
        self.target = target
        self.pca.fit(X_target)

        self.scores_target = self.pca.transform(X_target)
        self.loadings = self.pca.components_
        self.residuals_target = X_target - self.pca.inverse_transform(self.scores_target)

        self.Q_target = np.sum(self.residuals_target**2, axis=1)
        
        lambda_values = PCA().fit(X_target).explained_variance_[self.n_components:]
        theta1 = np.sum(lambda_values)
        theta2 = np.sum(lambda_values**2)
        theta3 = np.sum(lambda_values**3)
        h0 = 1 - 2 * theta1 * theta3 / (3 * theta2**2)
        calpha = scipy.stats.norm.ppf(self.q)
        fraction1 = calpha * np.sqrt(2 * theta2 * h0**2) / theta1
        fraction2 = theta2*h0*(h0-1) / theta1**2
        self.Q_conf = theta1 * (1 + fraction1 + fraction2)**(1/h0)

        self.Tsq_target = np.sum(
            (self.scores_target / np.std(self.scores_target, axis=0)) ** 2, axis=1
        )
        self.Tsq_conf = (
            f.ppf(q=0.95, dfn=self.n_components, dfd=self.scores_target.shape[0])
            * self.n_components
            * (self.scores_target.shape[0] - 1)
            / (self.scores_target.shape[0] - self.n_components)
        )

        self._fitted = True
        
        return self

    def predict(self, X_test, decision_rule="both"):
        """
        Applies fitted SIMCA model to test data and makes a prediction
        about the target class membership.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Test data as feature matrix.

        decision_rule : str, optional
            Prediction based on either 'Q', 'Tsq' or 'both',
            by default "both".

        Returns
        -------
        numpy.ndarray of shape (n_samples,)
            Boolean result array.

        Raises
        ------
        ValueError
            If invalid value is given for decision_rule argument.
        """
        # set as attributes for plot method
        
        self.scores_test = self.pca.transform(X_test)
        self.residuals_test = X_test - self.pca.inverse_transform(self.scores_test)
        self.Q_test = np.sum(self.residuals_test**2, axis=1)
        self.Tsq_test = np.sum(
            (self.scores_test / np.std(self.scores_test, axis=0)) ** 2, axis=1
        )

        pred_Q = self.Q_test < self.Q_conf
        pred_Tsq = self.Tsq_test < self.Tsq_conf
        self._validated = True

        if decision_rule == "both":
            return np.logical_and(pred_Q, pred_Tsq)
        elif decision_rule == "Q":
            return pred_Q
        elif decision_rule == "Tsq":
            return pred_Tsq
        else:
            raise ValueError("Invalid decision method, use 'Q', 'Tsq' or 'both'.")

    def Tsq_Q_plot(self, hue=None, annotate=None):
        """
        Visualizes fitted SIMCA model as T square Q plot.
        Test data is included when the 'predict' method was used prior.

        Parameters
        ----------
        hue : iterable, optional
            Iterable with test data class labels to color markers by class.
            Only labels for test data is needed, labels for target class are already known.
            Ignores if 'predict' method was not used prior,
            by default None.

        annotate : iterable, optional
            Iterable with sample names to annotate markers with,
            by default None.

        Returns
        -------
        matplotlib.pyplot.axes

        Raises
        ------
        ValueError
            If instance was not fitted before calling 'plot'.
        """
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        if self._validated:
            X = np.concatenate((self.Q_target, self.Q_test))
            Y = np.concatenate((self.Tsq_target, self.Tsq_test))
            style = ["Training"] * len(self.Q_target) + ["Validation"] * len(
                self.Q_test
            )
            if hue is not None:
                hue = [self.target] * len(self.Q_target) + list(hue)
        else:
            X = self.Q_target
            Y = self.Tsq_target
            style = ["Training"] * len(self.Q_target)
            hue = None

        ax = sns.scatterplot(x=X, y=Y, style=style, hue=hue)

        if annotate is not None:
            for x, y, name in zip(X, Y, annotate):
                ax.text(x, y, name)

        if self._validated:
            ax.legend()

        ax.axhline(self.Tsq_conf, c="tab:red", linestyle=":")
        ax.axvline(self.Q_conf, c="tab:red", linestyle=":")
        ax.set_xlabel("Q")
        ax.set_ylabel("$T^2$", rotation=0, labelpad=15)
        ax.set_title(f"SIMCA model for target: {self.target}")
        return ax
    
    def scores_plot(self, y_train, y_test, x_comp=1, y_comp=2):
        """
        Visualizes scores of fitted SIMCA model as Scores Plot.

        Parameters:
        ---------- 
        y_train : numpy.array of shape(n_samples,)
            True class labels for training data.    
        
        y_test : numpy.array of shape(n_samples,)
            True class labels for test data.
            
        x_comp : int, optional
            Component x axis, by default 1.

        y_comp : int, optional
            Component y axis, by default 2.
            
        Returns
        -------
        matplotlib.pyplot.axes         
        """        
        self.y_train = y_train
        self.y_test = y_test
        
        if not self._fitted:
            raise ValueError(
                "This model is not fitted yet! Call 'fit' with appropriate arguments before plotting."
            )

        if self._validated:
            X = np.concatenate(
                (self.scores_target[:, x_comp - 1], self.scores_test[:, x_comp - 1])
            )
            Y = np.concatenate(
                (self.scores_target[:, y_comp - 1], self.scores_test[:, y_comp - 1])
            )
            hue = list(self.y_train) + list(self.y_test)
            style = ["Training"] * len(self.y_train) + ["Validation"] * len(
                self.y_test
            )

        ax = sns.scatterplot(x=X, y=Y, hue=hue, style=style)

        ax.legend()

        plt.xlabel(f"PC {x_comp}")
        plt.ylabel(f"PC {y_comp}")

        return ax
    
    def plot_loadings(self, dataset, PC=1, color_range=0.1, width=6, height=6):
        """
        Plots loadings of a principle component with the original retention
        and drift time coordinates.

        Parameters
        ----------
        dataset : ims.Dataset
            The dataset is needed for the retention and drift time
            coordinates. 
        
        PC : int, optional
            principal component, by default 1.

        color_range : int, optional
            color_scale ranges from - color_range to + color_range
            centered at 0.

        width : int or float, optional
            plot width in inches, by default 9.

        height : int or float, optional
            plot height in inches, by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """

        # use retention and drift time axis from the first spectrum
        ret_time = dataset[0].ret_time
        drift_time = dataset[0].drift_time

        loading_pc = self.loadings[PC - 1, :].reshape(len(ret_time), len(drift_time))

        expl_var = []
        for i in range(1, self.n_components + 1):
            expl_var.append(round(self.pca.explained_variance_ratio_[i - 1] * 100, 1))

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            loading_pc,
            origin="lower",
            aspect="auto",
            cmap="RdBu_r",
            vmin=(-color_range),
            vmax=color_range,
            extent=(min(drift_time), max(drift_time), min(ret_time), max(ret_time)),
        )

        plt.colorbar().set_label("PCA loadings")

        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())

        plt.xlabel(dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        plt.title(f"PCA loadings of PC {PC} ({expl_var[PC-1]} % of variance)")
        return ax




class MultiClassSIMCA:
    """
    Builds soft independent modelling of class analogies model out off OneClassSIMCA models for each class.

    Parameters
    ----------
    models : list
        OneClassSIMCA model per class.

    Attributes
    ---------
    nclasses : int
        Number of OneClasSIMCA models in models.
        
    classnames : list
        Names of targets in OneClasSIMCA models.
        
    res_df : pd.DataFrame of shape (n_targets, n_samples)
        Binary representation of classification.
        Calculated in the predict method.

    Raises
    ------
    ValueError
        If instance was not fitted in 'OneClassSimca' class.

    Example
    -------
    >>> import numpy as np
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.utils import resample
    >>> from sklearn.metrics import accuracy_score
    >>>
    >>> # load example data
    >>> X, y = load_iris(return_X_y=True)
    >>>
    >>> # instantiate a model and fit target data
    >>> model = MultiClassSIMCA()
    >>> model.fit(X, y, n_components=2, q=0.95)
    >>>
    >>> # make prediction on test data and calculate accuracy
    >>> y_pred = model.predict(X_test)
    >>> y_true = y_test == target
    >>> print(accuracy_score(y_true, y_pred))
    >>>
    >>> # visualize results
    >>> model.plot()
    """
    
    def __init__(self, models=[]):
        self.models = models
        self.nclasses = len(models)
        self.classnames = [list(model.target) for model in models]
        for model in models:
            if not model._fitted:
                raise ValueError(
                f"OneClassSIMCA {model.target} is not fitted yet! Call 'fit' in OneClassSIMCA with appropriate arguments."
                ) 
            
    def fit(self, X_train, y_train, n_components=2, p=0.95):
        """
        Fit OneClasSIMCA models with training data.
        
        Parameters
        ----------
        X_train : numpy.ndarray of shape (n_samples, n_features)
            Training data as feature matrix.
        
        y_train : numpy.array of shape(n_samples,)
            True class labels for training data.
            
        n_components : int, optional
            Number of PCA components, by default 2.

        q : float, optional
            False discovery rate, by default 0.95.
        """
        
        self.models = []
        target_list = np.unique(y_train)
        for target in target_list:
            X_target = X_train[np.where(y_train == target)]
            model = OneClassSIMCA(n_components, p)
            model.fit(X_target, target)
            self.models.append(model)
            self.classnames.append(target)
            
        return self

    def predict(self, X_test, y_test=None):
        """
        Applies fitted SIMCA models to test data and makes a prediction
        about the class memberships.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Test data as feature matrix.
            
        y_test : numpy.array of shape(n_samples,), optional
            True class labels for test data, by default None

        Returns
        -------
        pd.DataFrame of shape (n_targets, n_samples)
        """
        
        y_pred_class = []
        for model in self.models:
            y_pred = model.predict(X_test)
            y_pred_class.append(y_pred.astype('int'))
            
        self.res_df = pd.DataFrame(
            data=np.array(y_pred_class).T,
            columns=[model.target for model in self.models]
        )
        
        if y_test is not None:
            self.res_df.insert(loc=0, column="Sample name", value=list(y_test))
            
        return self.res_df
    
    def plot(self):
        
        """Visualizes the classification of test data.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        
        self.res_df = self.res_df.set_index("Sample name").sort_index()
        
        # Define colors
        colors = ["tab:grey", "tab:cyan"]
        cmap = mcolors.LinearSegmentedColormap.from_list('Custom', colors, len(colors))

        ax = sns.heatmap(self.res_df, cmap=cmap)

        # Set the colorbar labels
        colorbar = ax.collections[0].colorbar
        colorbar.set_ticks([0.25,0.75])
        colorbar.set_ticklabels(['False', 'True'])
        
        ax.set(xlabel="Predicted label", ylabel="Sample name")
        
        return ax
    


import os
import re
import json
import h5py
import pywt
from array import array
from copy import deepcopy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from datetime import datetime
from time import ctime
from skimage.morphology import white_tophat, disk
from skimage.measure import label, regionprops
from skimage import measure
from zipfile import ZipFile
from ims.utils import asymcorr
from findpeaks import findpeaks
from scipy.signal import savgol_filter
from scipy import ndimage as ndi
from skimage.segmentation import watershed


class Spectrum:
    """
    Represents one GCIMS-Spectrum with the data matrix,
    retention and drift time coordinates.
    Sample or file name and timestamp are included unique identifiers.

    This class contains all methods that can be applied on a per spectrum basis,
    like I/O, plotting and some preprocessing tools.
    Methods that return a Spectrum change the instance inplace. Use the copy method.


    Parameters
    ----------
    name : str
        File or sample name as a unique identifier.
        Reader methods set this attribute to the file name without extension.

    values : numpy.array
        Intensity matrix.

    ret_time : numpy.array
        Retention time coordinate.

    drift_time : numpy.array
        Drift time coordinate.

    time : datetime object
        Timestamp when the spectrum was recorded.

    Example
    -------
    >>> import ims
    >>> sample = ims.Spectrum.read_mea("sample.mea")
    >>> sample.plot()
    """

    def __init__(self, name, values, ret_time, drift_time, time, meta_attr):
        self.name = name
        self.values = values
        self.ret_time = ret_time
        self.drift_time = drift_time
        self.time = time
        self.peak_table = None
        self._drift_time_label = "Drift time [ms]"
        self.meta_attr = meta_attr

    def __repr__(self):
        return f"GC-IMS Spectrum: {self.name}"

    # add, radd and truediv are implemented to calculate means easily
    def __add__(self, other):
        values = self.values + other.values
        ret_time = self.ret_time + other.ret_time
        drift_time = self.drift_time + other.drift_time
        x = Spectrum(self.name, values, ret_time, drift_time, self.time, self.meta_attr)
        x._drift_time_label = self._drift_time_label
        return x

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            raise NotImplementedError()

    def __truediv__(self, other):
        if isinstance(other, int):
            values = self.values / other
            ret_time = self.ret_time / other
            drift_time = self.drift_time / other
            x = Spectrum(self.name, values, ret_time, drift_time, self.time, self.meta_attr)
            x._drift_time_label = self._drift_time_label
            return x
        else:
            raise NotImplementedError()

    @property
    def shape(self):
        """
        Shape property of the data matrix.
        Equivalent to `ims.Spectrum.values.shape`.
        """
        return self.values.shape

    def copy(self):
        """
        Uses deepcopy from the copy module in the standard library.
        Most operations happen inplace. Use this method if you do not
        want to change the original variable.

        Returns
        -------
        Spectrum
            deepcopy of self.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> new_variable = sample.copy()
        """
        return deepcopy(self)

    @classmethod
    def read_zip(cls, path):
        """
        Reads zipped csv and json files from G.A.S Dortmund mea2zip converting tool.
        Present for backwards compatibility.
        Reading mea files is much faster and saves the manual extra step of converting.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_zip("sample.mea")
        >>> print(sample)
        GC-IMS Spectrum: sample
        """
        with ZipFile(path) as myzip:
            with myzip.open("csv_data.csv", "r") as mycsv:
                values = pd.read_csv(mycsv, header=None)
            with myzip.open("meta_attributes.json", "r") as myjson:
                meta_attr = json.load(myjson)

        values = np.array(values)
        values = np.delete(values, -1, axis=1)

        ret_time = (
            np.arange(meta_attr["Chunks count"])
            * (meta_attr["Chunk averages"] + 1)
            * meta_attr["Chunk trigger repetition [ms]"]
            / 1000
        )
        drift_time = (
            np.arange(meta_attr["Chunk sample count"])
            / meta_attr["Chunk sample rate [kHz]"]
        )

        path = os.path.normpath(path)
        name = os.path.split(path)[1]
        time = datetime.strptime(meta_attr["Timestamp"], "%Y-%m-%dT%H:%M:%S")

        return cls(name, values, ret_time, drift_time, time, meta_attr)

    @classmethod
    def read_mea(cls, path):
        """
        Reads mea files from G.A.S Dortmund instruments.
        Alternative constructor for ims.Spectrum class.
        Much faster than reading csv files and therefore preferred.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample)
        GC-IMS Spectrum: sample
        """
        path = os.path.normpath(path)
        name = os.path.split(path)[1]
        name = name.split(".")[0]

        with open(path, "rb") as f:
            content = f.read()
            i = content.index(0)
            meta_attr = content[: i - 1]
            meta_attr = meta_attr.decode("windows-1252")
            data = content[i + 1 :]
            data = array("h", data)

        meta_attr = meta_attr.split("\n")

        key_re = re.compile(r"^.*?(?==)")
        value_re = re.compile(r"(?<==)(.*?)(?=\[|$)")
        # unit_re = re.compile(r"\[(.*?)\]")

        for i in meta_attr:
            key = key_re.search(i).group(0).strip()
            value = value_re.search(i).group(0).strip()
            if "Chunks count" in key:
                chunks_count = int(value)
            elif "Chunk averages" in key:
                chunk_averages = int(value)
            elif "Chunk sample count" in key:
                chunk_sample_count = int(value)
            elif "Chunk sample rate" in key:
                chunk_sample_rate = int(value)
            elif "Chunk trigger repetition" in key:
                chunk_trigger_repetition = int(value)
            elif "Timestamp" in key:
                timestamp = datetime.strptime(value, '"%Y-%m-%dT%H:%M:%S"')

        data = np.array(data)
        data = data.reshape(chunks_count, chunk_sample_count)

        ret_time = (
            np.arange(chunks_count)
            * (chunk_averages + 1)
            * chunk_trigger_repetition
            / 1000
        )

        drift_time = np.arange(chunk_sample_count) / chunk_sample_rate

        return cls(name, data, ret_time, drift_time, timestamp, meta_attr)

    @classmethod
    def read_csv(cls, path):
        """
        Reads generic csv files. The first row must be
        the drift time values and the first column must be
        the retention time values. Values inbetween are the
        intensity matrix.
        Uses the time when the file was created as timestamp.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_csv("sample.csv")
        >>> print(sample)
        GC-IMS Spectrum: sample
        """
        name = os.path.split(path)[1]
        name = name.split(".")[0]
        df = pd.read_csv(path, index_col = [0])
        values = df.values
        ret_time = np.array(df.index)
        drift_time = df.columns.to_numpy().astype(float)
        timestamp = os.path.getctime(path)
        timestamp = ctime(timestamp)
        timestamp = datetime.strptime(timestamp, "%a %b  %d %H:%M:%S %Y")
        return cls(name, values, ret_time, drift_time, timestamp, meta_attr={})

    @classmethod
    def read_hdf5(cls, path):
        """
        Reads hdf5 files exported by the to_hdf5 method.
        Convenient way to store preprocessed spectra.
        Especially useful for larger datasets as preprocessing
        requires more time.
        Preferred to csv because of very fast read and write speeds.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.to_hdf5()
        >>> sample = ims.Spectrum.read_hdf5("sample.hdf5")
        """
        with h5py.File(path, "r") as f:
            values = np.array(f["values"])
            ret_time = np.array(f["ret_time"])
            drift_time = np.array(f["drift_time"])
            name = str(f.attrs["name"])
            time = datetime.strptime(f.attrs["time"], "%Y-%m-%dT%H:%M:%S")
            drift_time_label = str(f.attrs["drift_time_label"])
            if "meta_attr" in f.attrs:
                meta_attr = json.loads(f.attrs["meta_attr"])
            else:
                meta_attr = {}

        spectrum = cls(name, values, ret_time, drift_time, time, meta_attr)
        spectrum._drift_time_label = drift_time_label
        return spectrum

    def to_hdf5(self, path=None):
        """
        Exports spectrum as hdf5 file.
        Useful to save preprocessed spectra, especially for larger datasets.
        Preferred to csv format because of very fast read and write speeds.

        Parameters
        ----------
        path : str, optional
            Directory to export files to,
            by default the current working directory.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.to_hdf5()
        >>> sample = ims.Spectrum.read_hdf5("sample.hdf5")
        """
        if path is None:
            path = os.getcwd()

        with h5py.File(f"{path}/{self.name}.hdf5", "w-") as f:
            f.create_dataset("values", data=self.values)
            f.create_dataset("ret_time", data=self.ret_time)
            f.create_dataset("drift_time", data=self.drift_time)
            f.attrs["name"] = self.name
            f.attrs["time"] = datetime.strftime(self.time, "%Y-%m-%dT%H:%M:%S")
            f.attrs["drift_time_label"] = self._drift_time_label
            f.attrs["meta_attr"] = json.dumps(self.meta_attr)
    
    def normalization(self):
        """
        Normalize a single spectrum by scaling its values to the range [0, 1].

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.normalization()

        Returns
        -------
        self : Spectrum
            The spectrum with normalized values.
        """

        # Find the minimum and maximum values of the intensity values
        min_value = np.min(self.values)
        max_value = np.max(self.values)

        # Normalize the intensity values to the range [0, 1]
        if max_value > min_value:
            self.values = (self.values - min_value) / (max_value - min_value)
        # Raise Error
        else:
            raise ValueError(f"Spectrum {self.name} cannot be normalized because it has constant or zero values.")

        return self

    def find_peaks(self, limit=None, denoise="fastnl", window=30, verbose=0):
        """
        Automated GC-IMS peak detection based on persistent homology.

        Parameters
        ----------
        spectrum : ims.Spectrum
            GC-IMS spectrum to use.

        limit : float
            Values > limit are active search areas to detect regions of interest (ROI).
            If None limit is selected by the minimum persistence score,
            by default None.

        denoise : string, (default : 'fastnl', None to disable)
            Filtering method to remove noise:
                * None
                * 'fastnl'
                * 'bilateral'
                * 'lee'
                * 'lee_enhanced'
                * 'kuan'
                * 'frost'
                * 'median'
                * 'mean'

        window : int, (default : 30)
            Denoising window. Increasing the window size may removes noise better but may also removes details of image in certain denoising methods.

        verbose : int (default : 3)
            Print to screen. 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace.

        Returns
        -------
        pandas.DataFrame
            Peak table with drift and retention times,
            the correspondig x and y indices,
            the maximum intensity of the peak,
            birth and death levels and scores.

        References
        ----------
        Taskesen, E. (2020). findpeaks is for the detection of peaks and valleys in a 1D vector and 2D array (image). (Version 2.3.1) [Computer software]. https://erdogant.github.io/findpeaks
        """
        if limit is None:
            fp = findpeaks(
                method="topology",
                limit=0,
                scale=True,
                denoise=denoise,
                window=window,
                verbose=verbose,
            )
            fp.fit(self.values)
            limit = fp.results["persistence"]["score"].min()

        # actual peak detection
        fp = findpeaks(
            method="topology",
            limit=limit,
            scale=True,
            denoise=denoise,
            window=window,
            verbose=verbose,
        )
        fp.fit(self.values)

        # reindex to ensure consistent numbering and start at 1
        df = fp.results["persistence"].reset_index(drop=True)
        df.index = df.index + 1
        df.index.name = "peak number"

        df["x"] = df["x"].astype("int")
        df["y"] = df["y"].astype("int")

        # add the drift and retention time values for all indices
        if hasattr(self, "rip_ms"):
            df.insert(0, "riprel_dt", self.drift_time[df["x"].values])
            df.insert(3, "abs_dt", self.drift_time[df["x"].values] * self.rip_ms)
        else:
            df.insert(0, "abs_dt", self.drift_time[df["x"].values])
        df.insert(2, "ret_time", self.ret_time[df["y"].values])
        df.insert(0, "compound", "")
        df.insert(1, "intensity", self.values[df["y"].values, df["x"].values])

        self.peak_table = df
        return self

    def detect_peaks(self, threshold_rel=0.5, peak_size=10):
        """
        Fast peak detection using simple thresholding and connected components.
        Returns a labeled mask and a list of peak outlines. Make sure to cut out the RIP
        before using this method to avoid incorrect thresholding.

        Parameters
        ----------
        threshold_rel : float, default=0.5
            Relative threshold for peak detection. Decrease to be more sensitive, increase to to detect more intense peaks only.
        
        peak_size : int, default=10
            Minimum pixel peak size (number of connected pixels) required for a region to be considered a peak.
            Peaks with fewer pixels than this threshold will be filtered out as noise.

        Returns
        -------
        self : Spectrum
            The spectrum with updated peaklist attribute containing peak information

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.riprel().cut_dt().cut_rt()
        >>> sample.detect_peaks()
        >>> sample.plot_thresholding()
        """
        # Thresholding of 2d array
        threshold = threshold_rel * np.max(self.values)
        binary_mask = self.values > threshold

        # Label connected regions (peaks)
        labeled_mask = label(binary_mask)

        # Get outlines and peak information
        outlines = []
        peaks_data = []
        peak_counter = 1  # Counter for consistent peak labeling after filtering
        
        for region_label in range(1, labeled_mask.max() + 1):
            mask = labeled_mask == region_label
            
            # Get coordinates of the peak (indices where mask is True)
            coords = np.argwhere(mask)
            
            # Filter out small peaks based on peak_size
            if len(coords) < peak_size:
                continue
            
            contours = measure.find_contours(mask, 0.5)
            outlines.append(contours)
            
            # Find peak maximum within this region
            region_values = self.values * mask
            max_idx = np.unravel_index(np.argmax(region_values), region_values.shape)
            max_intensity = self.values[max_idx]
            
            # Calculate peak volume (background-subtracted)
            peak_volume = np.sum(self.values[mask] - threshold)
            
            # Get drift and retention time coordinates
            max_dt_idx = max_idx[1]  # x (drift time)
            max_rt_idx = max_idx[0]  # y (retention time)
            
            peak_entry = {
                'peak_label': peak_counter,
                'intensity': max_intensity,
                'volume': peak_volume,
                'ret_time': self.ret_time[max_rt_idx],
                'x': max_dt_idx,
                'y': max_rt_idx,
                'n_pixels': len(coords)
            }
            
           
            if hasattr(self, "rip_ms"):
                peak_entry['riprel_dt'] = self.drift_time[max_dt_idx]
                peak_entry['abs_dt'] = self.drift_time[max_dt_idx] * self.rip_ms
            else:
                peak_entry['abs_dt'] = self.drift_time[max_dt_idx]
                print("Warning! Make sure that you are using a region without the RIP, for peak detection to function properly.")
            peaks_data.append(peak_entry)
            peak_counter += 1  

        # create peak_list
        df = pd.DataFrame(peaks_data)
        
        if len(df) > 0:
            # Reorder columns to match find_peaks style
            if hasattr(self, "rip_ms"):
                column_order = ['peak_label', 'intensity', 'volume', 'riprel_dt', 'abs_dt', 'ret_time', 'x', 'y', 'n_pixels']
            else:
                column_order = ['peak_label', 'intensity', 'volume', 'abs_dt', 'ret_time', 'x', 'y', 'n_pixels']
            
            df = df[column_order]
            df.index = df.index + 1
            df.index.name = "peak number"
        
        # Store results in instanced dataframes
        self.peaklist = df
        print(f"Found {len(self.peaklist)} peaks")
        self.peaklist_mask = labeled_mask
        self.peaklist_outlines = outlines
        
        return self
    
    def plot_thresholding(self, outline_color="yellow", linewidth=2, annotate=False, **kwargs):
        """
        Plots GC-IMS spectrum with peak outlines from detect_peaks method.
        Must be called after detect_peaks method.

        Parameters
        ----------
        outline_color : str, default="yellow"
            Color for the peak outlines
        linewidth : float, default=2
            Width of the outline lines
        annotate : bool, default=False
            If True, display peak labels on the plot
        **kwargs
            Additional keyword arguments passed to the plot() method.
            
            - vmin : int, default=30
              Minimum intensity for the colormap
            - vmax : int, default=400
              Maximum intensity for the colormap
            - width : int, default=6
              Figure width in inches
            - height : int, default=6
              Figure height in inches

        Returns
        -------
        matplotlib.pyplot.axes
            The axes object with the plotted spectrum and peak outlines

        Example
        -------
        >>> sample.detect_peaks()
        >>> sample.plot_thresholding(annotate=True)
        >>> plt.show()
        """
        if not hasattr(self, 'peaklist') or self.peaklist is None:
            raise ValueError("Call 'detect_peaks' method first.")
        
        # Plot spectrum
        fig, ax = self.plot(**kwargs)
        
        # Overlay peak outlines using stored results
        for peak_contours in self.peaklist_outlines:
            for contour in peak_contours:
                y, x = contour.T
                ax.plot(
                    self.drift_time[x.astype(int)], 
                    self.ret_time[y.astype(int)], 
                    color=outline_color, 
                    linewidth=linewidth
                )
        
        # Add peak labels if requested
        if annotate:
            for i, row in self.peaklist.iterrows():
                if "riprel_dt" in self.peaklist.columns:
                    x = row["riprel_dt"]
                else:
                    x = row["abs_dt"]
                y = row["ret_time"]
                label = str(int(row["peak_label"]))
                ax.text(x, y, label, c="yellow", fontsize=12, ha='left', va='center')
        
        ax.set_title(f"{self.name} - {len(self.peaklist)} peaks detected")
        
        return ax

    def plot_peaks(self):
        """
        Plots GC-IMS spectrum with peak labels from findpeaks method.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if self.peak_table is None:
            raise ValueError("Call 'find_peaks' method first.")

        # call ims.Spectrum.plot method
        _, ax = self.plot()

        # iterate over peak table and add labels
        for i, row in self.peak_table.iterrows():
            if "riprel_dt" in self.peak_table.columns:
                x = row["riprel_dt"]
            else:
                x = row["abs_dt"]
            y = row["ret_time"]
            label = f"{i} {row['compound']}"
            ax.text(x, y, label, c="yellow", fontsize=12)

        return ax

    def plot_persistence(self):
        """
        Persistance plot of birth vs death levels from findpeak method.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if self.peak_table is None:
            raise ValueError("Call 'find_peaks' method first.")
        
        x = self.peak_table["birth_level"].values
        y = self.peak_table["death_level"].values

        ax = plt.subplot()
        ax.plot(x, y, ".", c="tab:blue")

        # draw line across plot
        X = np.c_[x, y]
        ax.plot([np.min(X), np.max(X)], [np.min(X), np.max(X)], "--", c="tab:grey")

        # set axis limits and labels
        ax.set_xlim((np.min(X), np.max(X)))
        ax.set_ylim((np.min(X), np.max(X)))
        ax.set_xlabel("Birth level")
        ax.set_ylabel("Death level")

        # add peak labels
        for i, row in self.peak_table.iterrows():
            x = row["birth_level"]
            y = row["death_level"]
            label = f"{i} {row['compound']}"
            ax.text(x, y, label)

        return ax

    def watershed_segmentation(self, threshold):
        """
        Finds boundaries for overlapping peaks using watershed segmentation.
        Requires peak_table for starting coordinates.

        Parameters
        ----------
        threshold : int
            Threshold is used to binarize the intensity values to calculate the distances.

        Returns
        -------
        numpy.ndarray
            Labels array with same shape as intensity values.
        """
        if self.peak_table is None:
            raise ValueError("Call 'find_peaks' method first.")
        
        # Binarize intensity values
        image = np.copy(self.values) >= threshold

        # Generate the markers as local maxima of the distance to the background
        distance = ndi.distance_transform_edt(image)

        # Use x and y coordinates from the peak table
        coords = self.peak_table[["y", "x"]].values

        # Watershed segmentation according to scikit-image tutorial
        mask = np.zeros(distance.shape, dtype=bool)
        mask[tuple(coords.T)] = True
        markers, _ = ndi.label(mask)
        labels = watershed(-distance, markers, mask=image)
        return labels

    def calc_reduced_mobility(self, T = 318.15, p = 1013.25, Ud = 2132, L = 5.3):
        """
        Calculates the reduced mobility values for the drift times denoted in the peak table.
        The formula for the calculation of the reduced mobility values originates from
        Ahrens, A., Zimmermann, S. Towards a hand-held, fast, and sensitive gas chromatograph-ion mobility spectrometer
        for detecting volatile compounds. Anal Bioanal Chem 413, 1009–1016 (2021). https://doi.org/10.1007/s00216-020-03059-9

        Parameters
        ----------
        T : float, optional
            Temperature of the IMS cell, by default 318.15
        p : float, optional
            pressure in the IMS cell, by default 1013.25
        Ud : int, optional
            Drift voltage, by default 2132
        L : float, optional
            Length of the drift tube, by default 5.3

        Returns
        -------
        pandas.DataFrame
            returns the original dataframe from the find_peaks method,
            but adds a column for the reduced mobility
        """
        if self.peak_table is None:
            raise ValueError("Call 'find_peaks' method first.")
        T0 = 273.15
        p0 = 1013.15

        dt = self.peak_table["abs_dt"].values
        K0 = (L**2 * T0 * p) / (dt * 10**-3 * Ud * T * p0)
        self.peak_table["mobility"] = K0
        return self

    def asymcorr(self, lam=1e7, p=1e-3, niter=20):
        """
        Retention time baseline correction using asymmetric least squares.

        Parameters
        ----------
        lam : float, optional
            Controls smoothness. Larger numbers return smoother curves,
            by default 1e7

        p : float, optional
            Controls asymmetry, by default 1e-3

        niter : int, optional
            Number of iterations during optimization,
            by default 20

        Returns
        -------
        Spectrum
        """
        for i in range(self.values.shape[1]):
            y = self.values[:, i]
            self.values[:, i] = asymcorr(y, lam=lam, p=p, niter=niter)

        return self

    def savgol(self, window_length=10, polyorder=2, direction="both"):
        """
        Applys a Savitzky-Golay filter to intensity values.
        Can be applied in the drift time, retention time or both directions.

        Parameters
        ----------
        window_length : int, optional
            The length of the filter window, by default 10

        polyorder : int, optional
            The order of the polynomial used to fit the samples, by default 2

        direction : str, optional
            The direction in which to apply the filter.
            Can be 'drift time', 'retention time' or 'both'.
            By default 'both'

        Returns
        -------
        Spectrum
        """
        if direction == "drift_time":
            axis = 1
        elif direction == "ret_time":
            axis = 0
        elif direction == "both":
            self.values = savgol_filter(self.values, window_length, polyorder, axis=0)
            axis = 1
        else:
            raise ValueError(
                "Only 'drift_time', 'ret_time' or 'both' are valid options!"
            )

        self.values = savgol_filter(self.values, window_length, polyorder, axis=axis)
        return self

    def tophat(self, size=15):
        """
        Applies white tophat filter on data matrix as a baseline correction.
        Size parameter is the diameter of the circular structuring element.
        (Slow with large size values.)

        Parameters
        ----------
        size : int, optional
            Size of structuring element, by default 15

        Returns
        -------
        Spectrum
        """
        self.values = white_tophat(self.values, disk(size))
        return self

    def sub_first_rows(self, n=1):
        """
        Subtracts first n rows from every row in spectrum.
        Effective and simple baseline correction
        if RIP tailing is a concern but can hide small peaks.

        Returns
        -------
        Spectrum
        """
        fl = self.values[0 : n - 1, :].mean(axis=0)
        self.values = self.values - fl
        return self

    def riprel(self):
        """
        Replaces drift time coordinate with RIP relative values.
        Useful to cut away the RIP because it´s position is set to 1.

        Does not interpolate the data matrix to a completly artificial
        axis like ims.Dataset.interp_riprel.

        Returns
        -------
        Spectrum
            RIP relative drift time coordinate
            otherwise unchanged.
        """
        
        rip_index = np.argmax(self.values, axis=1)[0]
        rip_ms = self.drift_time[rip_index]
        dt_riprel = self.drift_time / rip_ms
        self.drift_time = dt_riprel
        self._drift_time_label = "Drift time RIP relative"
        self.rip_ms = rip_ms
        return self

    def rip_scaling(self):
        """
        Scales values relative to global maximum.
        Can be useful to directly compare spectra from
        instruments with different sensitivity.

        Returns
        -------
        Spectrum
            With scaled values.
        """
        m = np.max(self.values)
        self.values = self.values / m
        return self

    def resample(self, n=2):
        """
        Resamples spectrum by calculating means of every n rows.
        If the length of the retention time is not divisible by n
        it and the data matrix get cropped by the remainder at the long end.

        Parameters
        ----------
        n : int, optional
            Number of rows to mean,
            by default 2.

        Returns
        -------
        Spectrum
            Resampled values.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.resample(2)
        >>> print(sample.shape)
        (2041, 3150)
        """
        a, _ = self.values.shape
        rest = a % n
        if rest != 0:
            self.values = self.values[: a - rest, :]
            self.ret_time = self.ret_time[: a - rest]

        self.values = (self.values[0::n, :] + self.values[1::n, :]) / n
        self.ret_time = self.ret_time[::n]
        return self

    def binning(self, n=2):
        """
        Downsamples spectrum by binning the array with factor n.
        Similar to ims.Spectrum.resampling but works on both dimensions
        simultaneously.
        If the dimensions are not divisible by the binning factor
        shortens it by the remainder at the long end.
        Very effective data reduction because a factor n=2 already
        reduces the number of features to a quarter.

        Parameters
        ----------
        n : int, optional
            Binning factor, by default 2.

        Returns
        -------
        Spectrum
            Downsampled data matrix.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.binning(2)
        >>> print(sample.shape)
        (2041, 1575)
        """
        a, b = self.values.shape
        rest0 = a % n
        rest1 = b % n

        if rest0 != 0:
            self.values = self.values[: a - rest0, :]
            self.ret_time = self.ret_time[: a - rest0]
        if rest1 != 0:
            self.values = self.values[:, : b - rest1]
            self.drift_time = self.drift_time[: b - rest1]

        new_dims = (a // n, b // n)

        shape = (new_dims[0], a // new_dims[0], new_dims[1], b // new_dims[1])

        self.values = self.values.reshape(shape).mean(axis=(-1, 1))
        self.ret_time = self.ret_time[::n]
        self.drift_time = self.drift_time[::n]
        return self
    
    def wavecompr(self, direction="ret_time", wavelet="db3", level=3):
        """
        Data reduction by wavelet compression.
        Can be applied to drift time, retention time or both axis.

        Parameters
        ----------
        direction : str, optional
            The direction in which to apply the filter.
            Can be 'drift time', 'retention time' or 'both'.
            By default 'ret_time'.

        wavelet : str, optional
            Wavelet object or name string,
            by default "db3".

        level : int, optional
            Decomposition level (must be >= 0),
            by default 3.

        Returns
        -------
        Spectrum

        Raises
        ------
        ValueError
            When direction is neither 'ret_time', 'drift_time' or 'both'.
        """
        if direction == "ret_time":
            coef_values = pywt.wavedec(
                self.values,
                wavelet=wavelet,
                level=level,
                axis=0
                )
            self.values = coef_values[0]
            self.ret_time = np.linspace(
                self.ret_time[0],
                stop=self.ret_time[-1],
                num=self.values.shape[0]
                )

        elif direction == "drift_time":
            coef_values = pywt.wavedec(
                self.values,
                wavelet=wavelet,
                level=level,
                axis=1
                )
            self.values = coef_values[0]
            self.drift_time = np.linspace(
                self.drift_time[0],
                stop=self.drift_time[-1],
                num=self.values.shape[1]
                )

        elif direction == "both":
            coef_values = pywt.wavedec2(
                self.values,
                wavelet=wavelet,
                level=level,
            )
            self.values = coef_values[0]
            self.ret_time = np.linspace(
                self.ret_time[0],
                stop=self.ret_time[-1],
                num=self.values.shape[0]
                )
            self.drift_time = np.linspace(
                self.drift_time[0],
                stop=self.drift_time[-1],
                num=self.values.shape[1]
                )

        else:
            raise ValueError("Direction must be 'ret_time', 'drift_time or 'both'!")

        return self

    def cut_dt(self, start, stop=None):
        """
        Cuts data along drift time coordinate.
        Range in between start and stop is kept.
        If stop is not given uses the end of the array instead.
        Combination with RIP relative drift time
        values makes it easier to cut the RIP away and focus
        on the peak area.

        Parameters
        ----------
        start : int or float
            Start value on drift time coordinate.

        stop : int or float, optional
            Stop value on drift time coordinate.
            If None uses the end of the array,
            by default None.

        Returns
        -------
        Spectrum
            New drift time range.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.riprel().cut_dt(1.05, 2)
        >>> print(sample.shape)
        (4082, 1005)
        """
        if stop is None:
            stop = len(self.drift_time)

        idx_start = np.abs(self.drift_time - start).argmin()
        idx_stop = np.abs(self.drift_time - stop).argmin()
        self.drift_time = self.drift_time[idx_start:idx_stop]
        self.values = self.values[:, idx_start:idx_stop]
        return self

    def cut_rt(self, start, stop=None):
        """
        Cuts data along retention time coordinate.
        Range in between start and stop is kept.
        If stop is not given uses the end of the array instead.

        Parameters
        ----------
        start : int or float
            Start value on retention time coordinate.

        stop : int or float, optional
            Stop value on retention time coordinate.
            If None uses the end of the array,
            by default None.

        Returns
        -------
        Spectrum
            New retention time range.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.cut_rt(80, 500)
        >>> print(sample.shape)
        (2857, 3150)
        """
        if stop is None:
            stop = len(self.ret_time)

        idx_start = np.abs(self.ret_time - start).argmin()
        idx_stop = np.abs(self.ret_time - stop).argmin()
        self.ret_time = self.ret_time[idx_start:idx_stop]
        self.values = self.values[idx_start:idx_stop, :]
        return self

    def plot(self, vmin=30, vmax=400, width=6, height=6):
        """
        Plots Spectrum using matplotlibs imshow.
        Use %matplotlib widget in IPython or %matplotlib notebook
        in jupyter notebooks to get an interactive plot widget.
        Returning the figure is needed to make the export plot utilities possible.

        Parameters
        ----------
        vmin : int, optional
            Minimum of color range, by default 30.

        vmax : int, optional
            Maximum of color range, by default 300.

        width : int, optional
            Width in inches, by default 9.

        height : int, optional
            Height in inches, by default 10.

        Returns
        -------
        tuple
            (matplotlib.figure.Figure, matplotlib.pyplot.axes)

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> fig, ax = sample.plot()
        """
        fig, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            self.values,
            origin="lower",
            aspect="auto",
            cmap="RdBu_r",
            vmin=vmin,
            vmax=vmax,
            extent=(
                min(self.drift_time),
                max(self.drift_time),
                min(self.ret_time),
                max(self.ret_time),
            ),
        )

        plt.colorbar().set_label("Intensities [arbitrary units]")
        plt.title(self.name)

        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())

        plt.xlabel(self._drift_time_label)
        plt.ylabel("Retention time [s]")

        return fig, ax

    def export_plot(self, path=None, dpi=300, file_format="jpg", **kwargs):
        """
        Saves the figure as image file. See the docs for
        matplotlib savefig function for supported file formats and kwargs
        (https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html).

        Parameters
        ----------
        path : str, optional
            Directory to save the image,
            by default current working directory.

        file_format : str, optional
            See matplotlib savefig docs for information about supported formats,
            by default 'jpg'.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.export_plot()
        """
        if path is None:
            path = os.getcwd()

        fig, _ = self.plot(**kwargs)
        fig.savefig(
            f"{path}/{self.name}.{file_format}",
            dpi=dpi,
            bbox_inches="tight",
            pad_inches=0.2,
        )
